<?php

namespace Kahanit\TableRateShipping\Model\Carrier;

// inheritance
use Magento\Shipping\Model\Carrier\AbstractCarrier;
use Magento\Shipping\Model\Carrier\CarrierInterface;
// injection
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Quote\Model\Quote\Address\RateResult\ErrorFactory;
use Psr\Log\LoggerInterface;
use Magento\Shipping\Model\Rate\ResultFactory;
use Magento\Quote\Model\Quote\Address\RateResult\MethodFactory;
use Kahanit\TableRateShipping\Model\CarrierTableRate;
use Kahanit\TableRateShipping\Model\CarrierTableRateMethod;
use Magento\Catalog\Model\ProductFactory;
use Magento\Checkout\Model\Session AS CheckoutSession;
// other
use Magento\Quote\Model\Quote\Address\RateRequest;
use Magento\Catalog\Model\Product\Type as ProductType;
use Magento\ConfigurableProduct\Model\Product\Type\Configurable as ConfigurableProductType;
use Magento\Store\Model\Store;

class TableRateShipping extends AbstractCarrier implements CarrierInterface
{
    protected $_code = 'tablerateshipping';

    protected $rateResultFactory;

    protected $rateMethodFactory;

    protected $carrierTableRate;

    protected $carrierTableRateMethod;

    protected $productFactory;

    protected $checkoutSession;

    public function __construct(
        ScopeConfigInterface $scopeConfig,
        ErrorFactory $rateErrorFactory,
        LoggerInterface $logger,
        ResultFactory $rateResultFactory,
        MethodFactory $rateMethodFactory,
        CarrierTableRate $carrierTableRate,
        CarrierTableRateMethod $carrierTableRateMethod,
        ProductFactory $productFactory,
        CheckoutSession $checkoutSession,
        array $data = []
    )
    {
        $this->rateResultFactory = $rateResultFactory;
        $this->rateMethodFactory = $rateMethodFactory;
        $this->carrierTableRate = $carrierTableRate;
        $this->carrierTableRateMethod = $carrierTableRateMethod;
        $this->productFactory = $productFactory;
        $this->checkoutSession = $checkoutSession;

        parent::__construct($scopeConfig, $rateErrorFactory, $logger, $data);
    }

    public function getAllowedMethods()
    {
        return [$this->getCarrierCode() => __($this->getConfigData('name'))];
    }

    public function collectRates(RateRequest $request)
    {
        if (!$this->isActive()) {
            return false;
        }

        // get all products ids to fetched more details
        $productIds = [];
        $quote = $this->checkoutSession->getQuote();
        $cartItems = $quote->getAllItems();
        foreach ($cartItems as $cartItem) {
            $itemProductId = $cartItem->getProduct()->getId();
            if (!in_array($itemProductId, $productIds)) {
                $productIds[] = $itemProductId;
            }
        }

        // get products and carriers with required attributes
        $storeId = $quote->getStoreId();
        $products = $this->productFactory->create()->getCollection()
            ->addFieldToSelect('weight')
            ->addFieldToSelect('price')
            ->addAttributeToSelect('volume_type')
            ->addAttributeToSelect('volume')
            ->joinTable(
                ['ctp' => $this->carrierTableRate->getResource()->getTable('carrier_table_rate_product')],
                'id_product = entity_id',
                ['GROUP_CONCAT(id_carrier SEPARATOR \',\') AS id_carriers'],
                ['id_shop' => [$storeId, Store::DEFAULT_STORE_ID]],
                'left'
            )
            ->addFieldToFilter('entity_id', ['IN' => $productIds])
            ->groupByAttribute('entity_id')
            ->getItems();
        $productsTemp = [];
        foreach ($products as $product) {
            $productsTemp[$product->getData('entity_id')] = $product->getData();
            if (isset($productsTemp[$product->getData('entity_id')]['volume_type'])) {
                $productsTemp[$product->getData('entity_id')]['volume_type']
                    = $product->getAttributeText('volume_type');
            }
        }
        $products = $productsTemp;
        unset($productsTemp);

        // prepare product details
        $productDetails = [];
        $cartItems = $quote->getAllVisibleItems();
        foreach ($cartItems as $cartItemKey => $cartItem) {
            $itemProductId = $cartItem->getProduct()->getId();
            $itemChildren = $cartItem->getChildren();
            $itemProductType = $cartItem->getProductType();
            $itemProductWeight = $cartItem->getWeight();
            $itemProductVolume = (isset($products[$itemProductId]['volume'])) ? $products[$itemProductId]['volume'] : 0;
            $itemProductQty = $cartItem->getQty();
            if (count($itemChildren) == 0) {
                $productDetails[$cartItemKey . '-' . $itemProductId] = [
                    'entity_id' => $itemProductId,
                    'weight' => $itemProductWeight * $itemProductQty,
                    'volume' => $itemProductVolume * $itemProductQty
                ];
            } else {
                if ($itemProductType == ProductType::TYPE_BUNDLE) {
                    $weight = 0;
                    if ($cartItem->getProduct()->getWeightType()) {
                        $weight = $itemProductWeight;
                    } else {
                        foreach ($itemChildren as $itemChild) {
                            $weight += $itemChild->getWeight() * $itemChild->getQty();
                        }
                    }
                    $volume = 0;
                    $itemProductVolumeType = (isset($products[$itemProductId]['volume_type'])) ? $products[$itemProductId]['volume_type'] : 'Fixed';
                    if (strcasecmp($itemProductVolumeType, 'Fixed') == 0) {
                        $volume = $itemProductVolume;
                    } else {
                        foreach ($itemChildren as $itemChild) {
                            $itemChildProductId = $itemChild->getProduct()->getId();
                            $itemChildVolume = (isset($products[$itemChildProductId]['volume'])) ? $products[$itemChildProductId]['volume'] : 0;
                            $volume += $itemChildVolume * $itemChild->getQty();
                        }
                    }
                    $productDetails[$cartItemKey . '-' . $itemProductId] = [
                        'entity_id' => $itemProductId,
                        'weight' => $weight * $itemProductQty,
                        'volume' => $volume * $itemProductQty
                    ];
                } elseif ($itemProductType == ConfigurableProductType::TYPE_CODE) {
                    $itemChild = array_shift($itemChildren);
                    $itemChildProductId = $itemChild->getProduct()->getId();
                    $itemChildVolume = (isset($products[$itemChildProductId]['volume'])) ? $products[$itemChildProductId]['volume'] : 0;
                    $productDetails[$cartItemKey . '-' . $itemProductId] = [
                        'entity_id' => $itemChildProductId,
                        'weight' => $itemProductWeight * $itemProductQty,
                        'volume' => $itemChildVolume * $itemProductQty
                    ];
                } else {
                    $productDetails[$cartItemKey . '-' . $itemProductId] = [
                        'entity_id' => $itemProductId,
                        'weight' => $itemProductWeight * $itemProductQty,
                        'volume' => $itemProductVolume * $itemProductQty
                    ];
                }
            }
            $productDetails[$cartItemKey . '-' . $itemProductId]['id_carriers']
                = $products[$productDetails[$cartItemKey . '-' . $itemProductId]['entity_id']]['id_carriers'];
            $productDetails[$cartItemKey . '-' . $itemProductId]['price'] = $cartItem->getRowTotalInclTax();
            $productDetails[$cartItemKey . '-' . $itemProductId]['ptprice'] = $cartItem->getRowTotal();
            $productDetails[$cartItemKey . '-' . $itemProductId]['qty'] = $cartItem->getQty();
        }
        $carriers = $this->carrierTableRateMethod->getResource()->getCarriers($storeId, true);
        $carriersTemp = [];
        foreach ($carriers as $carrier) {
            $carriersTemp[$carrier['id_carrier']] = $carrier;
        }
        $carriers = $carriersTemp;
        unset($carriersTemp);
        $customCarriers = $this->getCarriersForProducts($productDetails, $carriers);

        // prepare rate result
        $shippingAddress = $quote->getShippingAddress();
        $result = $this->rateResultFactory->create();
        foreach ($customCarriers as $customCarrierKey => $customCarrier) {
            $subCarriers = explode('-', $customCarrierKey);
            $vars = [];
            foreach ($subCarriers as $subCarrier) {
                $vars[$subCarrier] = [
                    'id_carrier' => $carriers[$subCarrier]['id_carrier'],
                    'carrier_name' => $carriers[$subCarrier]['carrier_name'],
                    'method_name' => $carriers[$subCarrier]['method_name'],
                    'tw' => 0,
                    'tp' => 0,
                    'tptp' => 0,
                    'tq' => 0,
                    'tv' => 0,
                    'skip' => -1,
                    'stop' => false
                ];
                $subCarrierProducts = $customCarrier['y-' . $subCarrier];
                foreach ($subCarrierProducts as $subCarrierProduct) {
                    $vars[$subCarrier]['tw'] = $vars[$subCarrier]['tw'] + $productDetails[$subCarrierProduct]['weight'];
                    $vars[$subCarrier]['tp'] = $vars[$subCarrier]['tp'] + $productDetails[$subCarrierProduct]['price'];
                    $vars[$subCarrier]['tptp'] = $vars[$subCarrier]['tptp'] + $productDetails[$subCarrierProduct]['ptprice'];
                    $vars[$subCarrier]['tq'] = $vars[$subCarrier]['tq'] + $productDetails[$subCarrierProduct]['qty'];
                    $vars[$subCarrier]['tv'] = $vars[$subCarrier]['tv'] + $productDetails[$subCarrierProduct]['volume'];
                }
                $vars[$subCarrier]['ts'] = $this->carrierTableRate->getResource()->getShippingPrice(
                    $subCarrier, $shippingAddress->getCountryId(), $shippingAddress->getRegionId(),
                    $shippingAddress->getCity(), $shippingAddress->getPostcode(), $vars[$subCarrier], $storeId
                );
            }

            $shippingPrice = array_column($vars, 'ts');
            foreach ($shippingPrice as $shippingPriceItem) {
                if ($shippingPriceItem === false) {
                    continue 2;
                }
            }
            $carrierName = implode(', ', array_column($vars, 'carrier_name'));
            $methodName = implode(', ', array_column($vars, 'method_name'));
            $carrierId = 'trs-' . implode('-', array_column($vars, 'id_carrier'));
            $shippingPrice = array_sum($shippingPrice);

            if ($shippingPrice != false) {
                $rateResultMethod = $this->rateMethodFactory->create();
                $rateResultMethod->setCarrier($this->getCarrierCode());
                $rateResultMethod->setCarrierTitle($carrierName);
                $rateResultMethod->setMethodTitle($methodName);
                $rateResultMethod->setMethod($carrierId);
                $rateResultMethod->setPrice($shippingPrice);
                $rateResultMethod->setCost($shippingPrice);
                $result->append($rateResultMethod);
            }
        }

        return $result;
    }

    public function getCarriersForProducts($products, $carriers)
    {
        $customCarriers = [];
        $shipsAllCarriers = [];
        $shipsParCarriers = [];
        foreach ($carriers as $carrierKey => $carrier) {
            if (!isset($customCarriers[$carrier['id_carrier']])) {
                $customCarriers[$carrier['id_carrier']] = [
                    'y' => [],
                    'n' => [],
                    'm' => []
                ];
            }
            foreach ($products as $productKey => $product) {
                $productCarriers = ($product['id_carriers'] != '') ? explode(',', $product['id_carriers']) : [];
                if (count($productCarriers) == 0 || in_array($carrier['id_carrier'], $productCarriers)) {
                    $customCarriers[$carrier['id_carrier']]['y'][] = $productKey;
                } else {
                    $customCarriers[$carrier['id_carrier']]['n'][] = $productKey;
                    $customCarriers[$carrier['id_carrier']]['p'][$productKey] = $product;
                }
            }
            if (count($customCarriers[$carrier['id_carrier']]['n']) == 0) {
                $shipsAllCarriers[$carrierKey] = $carrier;
            } elseif (count($customCarriers[$carrier['id_carrier']]['y']) > 0) {
                $shipsParCarriers[$carrierKey] = $carrier;
            } else {
                unset($customCarriers[$carrier['id_carrier']]);
            }
        }
        if (count($shipsAllCarriers) > 0) {
            $returnCarriers = [];
            foreach ($shipsAllCarriers as $shipsAllCarrierKey => $shipsAllCarrier) {
                $returnCarriers[$shipsAllCarrier['id_carrier']] = $customCarriers[$shipsAllCarrier['id_carrier']];
                $returnCarriers[$shipsAllCarrier['id_carrier']]['y-' . $shipsAllCarrier['id_carrier']] = $returnCarriers[$shipsAllCarrier['id_carrier']]['y'];
                unset($returnCarriers[$shipsAllCarrier['id_carrier']]['y']);
                unset($returnCarriers[$shipsAllCarrier['id_carrier']]['n']);
                unset($returnCarriers[$shipsAllCarrier['id_carrier']]['m']);
                unset($returnCarriers[$shipsAllCarrier['id_carrier']]['p']);
            }
            return $returnCarriers;
        } else {
            $searchCarriers = $carriers;
            foreach ($shipsParCarriers as $shipsParCarrierKey => $shipsParCarrier) {
                unset($searchCarriers[$shipsParCarrierKey]);
                $carrierProducts = $this->getCarriersForProducts($customCarriers[$shipsParCarrier['id_carrier']]['p'], $searchCarriers);
                unset($customCarriers[$shipsParCarrier['id_carrier']]['p']);
                foreach ($carrierProducts as $carrierProductKey => $carrierProduct) {
                    $carrierProduct['y-' . $shipsParCarrier['id_carrier']] = $customCarriers[$shipsParCarrier['id_carrier']]['y'];
                    $customCarriers[$shipsParCarrier['id_carrier'] . '-' . $carrierProductKey] = $carrierProduct;
                }
                unset($customCarriers[$shipsParCarrier['id_carrier']]);
            }
            return $customCarriers;
        }
    }
}
