<?php

namespace Kahanit\TableRateShipping\Model\ResourceModel;

// inheritance
use Magento\Framework\Model\ResourceModel\Db\AbstractDb;
// injection
use Magento\Framework\Model\ResourceModel\Db\Context;
use Magento\Framework\FileSystem;
use Magento\Directory\Model\CountryFactory;
use Magento\Directory\Model\RegionFactory;
use Kahanit\TableRateShipping\Model\CarrierTableRateFactory;
use Kahanit\TableRateShipping\Model\CarrierTableRateZoneFactory;
// other
use Magento\Framework\App\Filesystem\DirectoryList;
use Kahanit\TableRateShipping\Helper\TRSHelper;
use Kahanit\TableRateShipping\Helper\TRSCSV;
use Symfony\Component\ExpressionLanguage\ExpressionLanguage;
use Kahanit\TableRateShipping\Helper\TRSExpressionFunctionProvider;
use Magento\Store\Model\Store;

class CarrierTableRate extends AbstractDb
{
    protected $fileSystem;

    protected $carrierTableRateZoneFactory;

    protected $countryFactory;

    protected $regionFactory;

    protected $carrierTableRateFactory;

    public function __construct(
        Context $context,
        FileSystem $fileSystem,
        CountryFactory $countryFactory,
        RegionFactory $regionFactory,
        CarrierTableRateFactory $carrierTableRateFactory,
        CarrierTableRateZoneFactory $carrierTableRateZoneFactory,
        $connectionName = null
    )
    {
        $this->fileSystem = $fileSystem;
        $this->countryFactory = $countryFactory;
        $this->regionFactory = $regionFactory;
        $this->carrierTableRateFactory = $carrierTableRateFactory;
        $this->carrierTableRateZoneFactory = $carrierTableRateZoneFactory;

        parent::__construct($context, $connectionName);
    }

    protected function _construct()
    {
        $this->_init('carrier_table_rate', 'id_carrier_table_rate');
    }

    /* functions group one */

    public function validateFields(\Kahanit\TableRateShipping\Model\CarrierTableRate $carrierTableRate)
    {
        /* validate zone */
        $zone = $this->carrierTableRateZoneFactory->create()->load($carrierTableRate->getData('id_zone'));
        if ($carrierTableRate->getData('id_zone') != 0 && $zone->isEmpty()) {
            return 'Invalid `Zone`';
        }

        /* validate country */
        $country = $this->countryFactory->create()->load($carrierTableRate->getData('id_country'));
        if ($carrierTableRate->getData('id_country') != '0' && $country->isEmpty()) {
            return 'Invalid `Country`';
        }

        /* validate state */
        $state = $this->regionFactory->create()->load($carrierTableRate->getData('id_state'));
        if ($carrierTableRate->getData('id_state') != 0 && $state->isEmpty()) {
            return 'Invalid `State`';
        }

        /* validate city */
        if (!preg_match('/^[^!<>;?=+@#"°{}_$%]*$/u', $carrierTableRate->getData('dest_city'))) {
            return 'Invalid `City`';
        }

        /* validate zip codes */
        $dest_zip = explode('-', $carrierTableRate->getData('dest_zip'));
        $dest_zip_from = (isset($dest_zip[0])) ? trim($dest_zip[0]) : '';
        $dest_zip_from = ($dest_zip_from == '*') ? '' : $dest_zip_from;
        $dest_zip_to = (isset($dest_zip[1])) ? trim($dest_zip[1]) : '';
        $dest_zip_to = ($dest_zip_to == '*') ? '' : $dest_zip_to;
        $dest_zip_to = ($dest_zip_to == '') ? $dest_zip_from : $dest_zip_to;
        if (strcasecmp($dest_zip_from, $dest_zip_to) === 0) {
            $carrierTableRate->setDestZip($dest_zip_from);
        } else {
            $carrierTableRate->setDestZip($dest_zip_from . '-' . $dest_zip_to);
        }

        /* validate zip/post code */
        if (!preg_match('/^[0-9a-zA-Z\-\s\*%_]*$/', $dest_zip_from)
            || !preg_match('/^[0-9a-zA-Z\-\s\*%_]*$/', $dest_zip_to)
        ) {
            return 'Invalid `Zip/Post code`';
        }

        /* validate zip from and to range */
        if ($dest_zip_from != '' && $dest_zip_to != '' && $dest_zip_to < $dest_zip_from) {
            return '`Zip/Post code from` should be less than `Zip/Post code to`';
        }

        /* validate to cannot specify range if wildcard is used */
        if ($dest_zip_from != '' && $dest_zip_to != '' && strcasecmp($dest_zip_from, $dest_zip_to) != 0
            && ((strpos($dest_zip_from, '*') != false || strpos($dest_zip_to, '*') != false)
                || (strpos($dest_zip_from, '%') != false || strpos($dest_zip_to, '%') != false)
                || (strpos($dest_zip_from, '_') != false || strpos($dest_zip_to, '_') != false))
        ) {
            return 'You cannot specify range if wildcard is used in zip/post code';
        }

        /* validate condition value from and to range */
        $condition_error_message = [
            'condition_weight_from_less_than_condition_weight_to' => '`Weight from` should be less than `Weight to`',
            'condition_price_from_less_than_condition_price_to' => '`Price from` should be less than `Price to`',
            'condition_ptprice_from_less_than_condition_ptprice_to' => '`Price(pretax) from` should be less than `Price(pretax) to`',
            'condition_quantity_from_less_than_condition_quantity_to' => '`Quantity from` should be less than `Quantity to`',
            'condition_volume_from_less_than_condition_volume_to' => '`Volume from` should be less than `Volume to`'
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($condition_types as $condition_type) {
            if ($carrierTableRate->getData('condition_' . $condition_type . '_to') < $carrierTableRate->getData('condition_' . $condition_type . '_from')) {
                return $condition_error_message['condition_' . $condition_type . '_from_less_than_condition_' . $condition_type . '_to'];
            }
        }

        /* validate symfony express language component installed */
        if (!class_exists('Symfony\\Component\\ExpressionLanguage\\ExpressionLanguage')) {
            return 'Symfony expression language component not found';
        }

        /* validate price */
        $vars = [
            'tw' => 1,
            'tp' => 1,
            'tptp' => 1,
            'tq' => 1,
            'tv' => 1,
            'ctwf' => 1,
            'ctwt' => 1,
            'ctwi' => 1,
            'ctpf' => 1,
            'ctpt' => 1,
            'ctpi' => 1,
            'ctptpf' => 1,
            'ctptpt' => 1,
            'ctptpi' => 1,
            'ctqf' => 1,
            'ctqt' => 1,
            'ctqi' => 1,
            'ctvf' => 1,
            'ctvt' => 1,
            'ctvi' => 1,
            'skip' => -1,
            'stop' => false
        ];
        $carrierTableRate->setData('price', str_replace('$', '', $carrierTableRate->getData('price'))); // for bc with old formula variables with $
        $language = new ExpressionLanguage();
        $language->registerProvider(new TRSExpressionFunctionProvider());
        try {
            $language->evaluate($carrierTableRate->getData('price'), $vars);
        } catch (\Exception $e) {
            return 'Invalid `Price`';
        }

        /* validate duplicate entry */
        $rule_exists = $this->ruleExists($carrierTableRate->getIdCarrier(), $carrierTableRate->getData(), $carrierTableRate->getIdShop());
        if ($rule_exists !== 0 && $rule_exists !== $carrierTableRate->getId()) {
            return 'Duplicate table rate found';
        }

        return (int)true;
    }

    public function ruleExists($id_carrier, $record, $id_shop)
    {
        $connection = $this->getConnection();
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['ct.id_carrier_table_rate']
        );
        $sql->where(
            '   ct.id_carrier = ' . (int)$id_carrier . '
            AND ct.id_zone = ' . (int)$record['id_zone'] . '
            AND ct.id_country = \'' . TRSHelper::addslashes($record['id_country']) . '\'
            AND ct.id_state = ' . (int)$record['id_state'] . '
            AND ct.dest_city = \'' . TRSHelper::addslashes($record['dest_city']) . '\'
            AND ct.dest_zip = \'' . TRSHelper::addslashes($record['dest_zip']) . '\'
            AND ct.condition_weight_from = ' . (float)$record['condition_weight_from'] . '
            AND ct.condition_weight_to = ' . (float)$record['condition_weight_to'] . '
            AND ct.condition_price_from = ' . (float)$record['condition_price_from'] . '
            AND ct.condition_price_to = ' . (float)$record['condition_price_to'] . '
            AND ct.condition_ptprice_from = ' . (float)$record['condition_ptprice_from'] . '
            AND ct.condition_ptprice_to = ' . (float)$record['condition_ptprice_to'] . '
            AND ct.condition_quantity_from = ' . (float)$record['condition_quantity_from'] . '
            AND ct.condition_quantity_to = ' . (float)$record['condition_quantity_to'] . '
            AND ct.condition_volume_from = ' . (float)$record['condition_volume_from'] . '
            AND ct.condition_volume_to = ' . (float)$record['condition_volume_to'] . '
            AND ct.price = \'' . TRSHelper::addslashes($record['price']) . '\'
            AND ct.id_shop = ' . (int)$id_shop
        );

        return (int)$connection->fetchOne($sql);
    }

    /* functions group two */

    public function getRuleGroups($id_carrier, $id_zone = null, $start = null, $length = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        // sub query
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['MIN(ct.id_carrier_table_rate) AS id_carrier_table_rate', 'ct.id_zone', 'ct.id_country', 'ct.id_state', 'ct.dest_city', 'ct.dest_zip',
                'GROUP_CONCAT(ct.condition_weight_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_weight_from',
                'GROUP_CONCAT(ct.condition_weight_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_weight_to',
                'GROUP_CONCAT(ct.condition_price_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_price_from',
                'GROUP_CONCAT(ct.condition_price_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_price_to',
                'GROUP_CONCAT(ct.condition_ptprice_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_ptprice_from',
                'GROUP_CONCAT(ct.condition_ptprice_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_ptprice_to',
                'GROUP_CONCAT(ct.condition_quantity_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_quantity_from',
                'GROUP_CONCAT(ct.condition_quantity_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_quantity_to',
                'GROUP_CONCAT(ct.condition_volume_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_volume_from',
                'GROUP_CONCAT(ct.condition_volume_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_volume_to',
                'GROUP_CONCAT(ct.price ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS price',
                'GROUP_CONCAT(ct.active ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS active',
                'MIN(ct.order) AS order', 'z.name AS zone']
        );
        $sql->joinLeft(['z' => $this->getTable('carrier_table_rate_zone')], 'z.id_zone = ct.id_zone AND z.id_shop = ' . (int)$id_shop, []);
        $sql->where(
            'ct.id_carrier = ' . (int)$id_carrier
            . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : '')
            . (($id_shop !== false) ? ' AND ct.id_shop = ' . (int)$id_shop : '')
        );
        $sql->group(['ct.id_country', 'ct.id_state', 'ct.dest_city', 'ct.dest_zip']);
        $sql->order(['MIN(ct.order) ASC', 'MIN(ct.id_carrier_table_rate) ASC']);

        // outer query
        $sql_outer = $connection->select();
        $sql_outer->from(
            ['st' => $sql],
            ['st.id_zone', 'st.zone', 'st.id_country', 'st.id_state', 'st.dest_city',
                'GROUP_CONCAT(st.dest_zip ORDER BY st.order ASC, st.id_carrier_table_rate ASC SEPARATOR \',\') AS dest_zip']
        );
        $sql_outer->group(
            ['st.id_country', 'st.id_state', 'st.dest_city',
                'st.condition_weight_from', 'st.condition_weight_to',
                'st.condition_price_from', 'st.condition_price_to',
                'st.condition_ptprice_from', 'st.condition_ptprice_to',
                'st.condition_quantity_from', 'st.condition_quantity_to',
                'st.condition_volume_from', 'st.condition_volume_to',
                'st.price', 'st.active']
        );
        $sql_outer->order(['MIN(st.order) ASC', 'MIN(st.id_carrier_table_rate) ASC']);
        if ($start !== null && $length !== null) {
            $sql_outer->limit($length, $start);
        }

        $results = $connection->fetchAll($sql_outer);

        // get country names from id
        $country_id = array_unique(array_column($results, 'id_country'));
        $country = [];
        if (count($country_id) > 0) {
            $country_collection = $this->countryFactory->create()->getCollection();
            $country_collection->addFieldToFilter('country_id', ['IN', $country_id]);
            foreach ($country_collection->getItems() as $key => $value) {
                $country[$key] = [
                    'iso_code' => $value->getData('iso2_code'),
                    'name' => $value->getName()
                ];
            }
        }

        // get state names from id
        $state_id = array_unique(array_column($results, 'id_state'));
        $state = [];
        if (count($state_id) > 0) {
            $state_collection = $this->regionFactory->create()->getCollection();
            $state_collection->addFieldToFilter('main_table.region_id', ['IN', $state_id]);
            foreach ($state_collection->getItems() as $key => $value) {
                $state[$key] = [
                    'iso_code' => $value->getData('code'),
                    'name' => $value->getName()
                ];
            }
        }

        foreach ($results as &$result) {
            $result['id_zone'] = ($result['id_zone'] == 0) ? '*' : $result['id_zone'];
            $result['zone'] = ($result['zone'] == '') ? '*' : $result['zone'];
            $result['id_country'] = ($result['id_country'] == '0') ? '*' : $result['id_country'];
            $result['c_iso_code'] = ($result['id_country'] == '0') ? '*' : ((isset($country[$result['id_country']])) ? $country[$result['id_country']]['iso_code'] : $result['id_country']);
            $result['country'] = ($result['id_country'] == '0') ? '*' : ((isset($country[$result['id_country']])) ? $country[$result['id_country']]['name'] : $result['id_country']);
            $result['id_state'] = ($result['id_state'] == 0) ? '*' : $result['id_state'];
            $result['s_iso_code'] = ($result['id_state'] == 0) ? '*' : ((isset($state[$result['id_state']])) ? $state[$result['id_state']]['iso_code'] : $result['id_state']);
            $result['state'] = ($result['id_state'] == 0) ? '*' : ((isset($state[$result['id_state']])) ? $state[$result['id_state']]['name'] : $result['id_state']);
            $result['dest_city'] = ($result['dest_city'] == '') ? '*' : TRSHelper::stripslashes($result['dest_city']);
            $result['dest_zip'] = ($result['dest_zip'] == '') ? '*' : TRSHelper::stripslashes($result['dest_zip']);
        }

        return $results;
    }

    public function getRuleGroupsNumRows($id_carrier, $id_zone = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        // sub query
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['MIN(ct.id_carrier_table_rate) AS id_carrier_table_rate', 'ct.id_zone', 'ct.id_country', 'ct.id_state', 'ct.dest_city', 'ct.dest_zip',
                'GROUP_CONCAT(ct.condition_weight_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_weight_from',
                'GROUP_CONCAT(ct.condition_weight_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_weight_to',
                'GROUP_CONCAT(ct.condition_price_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_price_from',
                'GROUP_CONCAT(ct.condition_price_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_price_to',
                'GROUP_CONCAT(ct.condition_ptprice_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_ptprice_from',
                'GROUP_CONCAT(ct.condition_ptprice_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_ptprice_to',
                'GROUP_CONCAT(ct.condition_quantity_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_quantity_from',
                'GROUP_CONCAT(ct.condition_quantity_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_quantity_to',
                'GROUP_CONCAT(ct.condition_volume_from ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_volume_from',
                'GROUP_CONCAT(ct.condition_volume_to ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS condition_volume_to',
                'GROUP_CONCAT(ct.price ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS price',
                'GROUP_CONCAT(ct.active ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS active',
                'MIN(ct.order) AS order']
        );
        $sql->where(
            'ct.id_carrier = ' . (int)$id_carrier
            . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : '')
            . (($id_shop !== false) ? ' AND ct.id_shop = ' . (int)$id_shop : '')
        );
        $sql->group(['ct.id_country', 'ct.id_state', 'ct.dest_city', 'ct.dest_zip']);
        $sql->order(['MIN(ct.order) ASC', 'MIN(ct.id_carrier_table_rate) ASC']);

        // outer query
        $sql_outer = $connection->select();
        $sql_outer->from(
            ['st' => $sql],
            ['st.id_zone', 'st.id_country', 'st.id_state', 'st.dest_city',
                'GROUP_CONCAT(st.dest_zip ORDER BY st.order ASC, st.id_carrier_table_rate ASC SEPARATOR \',\') AS dest_zip']
        );
        $sql_outer->group(
            ['st.id_country', 'st.id_state', 'st.dest_city',
                'st.condition_weight_from', 'st.condition_weight_to',
                'st.condition_price_from', 'st.condition_price_to',
                'st.condition_ptprice_from', 'st.condition_ptprice_to',
                'st.condition_quantity_from', 'st.condition_quantity_to',
                'st.condition_volume_from', 'st.condition_volume_to',
                'st.price', 'st.active']
        );
        $sql_outer->order(['MIN(st.order) ASC', 'MIN(st.id_carrier_table_rate) ASC']);

        // count query
        $sql_count = 'SELECT COUNT(*) AS total FROM (' . $sql_outer . ') AS tbl';

        return (int)$connection->fetchOne($sql_count);
    }

    public function getRules($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $dest_zip, $start = null, $length = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['ct.*', 'GROUP_CONCAT(ct.id_carrier_table_rate ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS id_carrier_table_rate',
                'GROUP_CONCAT(ct.dest_zip ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS dest_zip']
        );
        $sql->where(
            '   ct.id_carrier = ' . (int)$id_carrier . '
            AND ct.id_zone = ' . (int)$id_zone . '
            AND ct.id_country = \'' . TRSHelper::addslashes((($id_country == '*' || $id_country == '') ? '0' : $id_country), true) . '\'
            AND ct.id_state = ' . (int)(($id_state == '*' || $id_state == '') ? 0 : $id_state) . '
            AND ct.dest_city = \'' . TRSHelper::addslashes((($dest_city == '*' || $dest_city == '') ? '' : $dest_city), true) . '\'
            ' . (($dest_zip == '*' || $dest_zip == '')
                ? 'AND ct.dest_zip = \'\''
                : 'AND FIND_IN_SET(ct.dest_zip, \'' . TRSHelper::addslashes($dest_zip, true) . '\')') . '
            AND ct.id_shop = ' . (int)$id_shop
        );
        $sql->group(
            ['ct.condition_weight_from', 'ct.condition_weight_to',
                'ct.condition_price_from', 'ct.condition_price_to',
                'ct.condition_ptprice_from', 'ct.condition_ptprice_to',
                'ct.condition_quantity_from', 'ct.condition_quantity_to',
                'ct.condition_volume_from', 'ct.condition_volume_to',
                'ct.price', 'ct.active']
        );
        $sql->order(['MIN(ct.order) ASC', 'MIN(ct.id_carrier_table_rate) ASC']);
        if ($start !== null && $length !== null) {
            $sql->limit($length, $start);
        }

        $results = $connection->fetchAll($sql);
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($results as &$result) {
            $result['id_country'] = ($result['id_country'] == '0') ? '*' : $result['id_country'];
            $result['id_state'] = ($result['id_state'] == 0) ? '*' : $result['id_state'];
            $result['dest_city'] = ($result['dest_city'] == '') ? '*' : TRSHelper::stripslashes($result['dest_city']);
            $result['dest_zip'] = ($result['dest_zip'] == '') ? '*' : TRSHelper::stripslashes($result['dest_zip']);
            foreach ($condition_types as $condition_type) {
                $result['condition_' . $condition_type . '_from'] = ($result['condition_' . $condition_type . '_from'] == -1) ? '*' : $result['condition_' . $condition_type . '_from'];
                $result['condition_' . $condition_type . '_to'] = ($result['condition_' . $condition_type . '_to'] == -1) ? '*' : $result['condition_' . $condition_type . '_to'];
            }
            $result['price'] = TRSHelper::stripslashes($result['price']);
            $result['comment'] = TRSHelper::stripslashes($result['comment']);
        }

        return $results;
    }

    public function getRulesNumRows($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $dest_zip, $where, $id_shop)
    {
        $connection = $this->getConnection();

        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['GROUP_CONCAT(ct.id_carrier_table_rate ORDER BY ct.order ASC, ct.id_carrier_table_rate ASC SEPARATOR \',\') AS id_carrier_table_rate']
        );
        $sql->where(
            '   ct.id_carrier = ' . (int)$id_carrier . '
            AND ct.id_zone = ' . (int)$id_zone . '
            AND ct.id_country = \'' . TRSHelper::addslashes((($id_country == '*' || $id_country == '') ? '0' : $id_country), true) . '\'
            AND ct.id_state = ' . (int)(($id_state == '*' || $id_state == '') ? 0 : $id_state) . '
            AND ct.dest_city = \'' . TRSHelper::addslashes((($dest_city == '*' || $dest_city == '') ? '' : $dest_city), true) . '\'
            ' . (($dest_zip == '*' || $dest_zip == '')
                ? 'AND ct.dest_zip = \'\''
                : 'AND FIND_IN_SET(ct.dest_zip, \'' . TRSHelper::addslashes($dest_zip, true) . '\')') . '
            AND ct.id_shop = ' . (int)$id_shop
        );
        $sql->group(
            ['ct.condition_weight_from', 'ct.condition_weight_to',
                'ct.condition_price_from', 'ct.condition_price_to',
                'ct.condition_ptprice_from', 'ct.condition_ptprice_to',
                'ct.condition_quantity_from', 'ct.condition_quantity_to',
                'ct.condition_volume_from', 'ct.condition_volume_to',
                'ct.price', 'ct.active']
        );
        $sql->order(['MIN(ct.order) ASC', 'MIN(ct.id_carrier_table_rate) ASC']);
        $sql_count = 'SELECT COUNT(id_carrier_table_rate) AS total FROM (' . $sql . ') AS tbl';

        return (int)$connection->fetchOne($sql_count);
    }

    /* functions group three */

    public function saveRules($action, $id_carrier, $records, $id_shop)
    {
        $connection = $this->getConnection();
        $dir_var = $this->fileSystem->getDirectoryWrite(DirectoryList::VAR_DIR);
        if (!$dir_var->isFile('import/Kahanit/TableRateShipping/ImportStatus.json')) {
            $dir_var->writeFile('import/Kahanit/TableRateShipping/ImportStatus.json', '{"rowImported":0}');
        }

        $status = [
            'error' => 0,
            'success' => 0,
            'status' => '',
            'message' => []
        ];
        $fields = [
            'id_shop' => 0,
            'id_carrier' => 0,
            'id_zone' => 0,
            'id_country' => '0',
            'id_state' => 0,
            'dest_city' => '',
            'dest_zip' => '',
            'condition_weight_from' => -1,
            'condition_weight_to' => -1,
            'condition_price_from' => -1,
            'condition_price_to' => -1,
            'condition_ptprice_from' => -1,
            'condition_ptprice_to' => -1,
            'condition_quantity_from' => -1,
            'condition_quantity_to' => -1,
            'condition_volume_from' => -1,
            'condition_volume_to' => -1,
            'price' => '',
            'comment' => '',
            'active' => 0,
            'order' => 0
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        $insert_carrier_table_rates = [];
        $insert_carrier_table_rates_flat = [];

        if (is_array($records) && count($records) > 0) {
            foreach ($records as $record_index => &$record) {
                // assign shop and carrier
                $record['id_shop'] = $id_shop;
                $record['id_carrier'] = $id_carrier;

                // prepare condition from and to values
                foreach ($condition_types as $condition_type) {
                    if (isset($record['condition_' . $condition_type])) {
                        if (is_numeric($record['condition_' . $condition_type])) {
                            $record['condition_' . $condition_type . '_from'] = (float)$record['condition_' . $condition_type];
                            $record['condition_' . $condition_type . '_to'] = (float)$record['condition_' . $condition_type];
                        } else {
                            $condition_values = explode('-', $record['condition_' . $condition_type]);
                            if (isset($condition_values[0]) && is_numeric($condition_values[0]) && isset($condition_values[1]) && is_numeric($condition_values[1])) {
                                $record['condition_' . $condition_type . '_from'] = (float)$condition_values[0];
                                $record['condition_' . $condition_type . '_to'] = (float)$condition_values[1];
                            }
                        }
                        unset($record['condition_' . $condition_type]);
                    }
                }

                // prepare record
                foreach ($fields as $field_key => $field_default) {
                    if (isset($record[$field_key])) {
                        $record[$field_key] = (is_int($field_default))
                            ? $record[$field_key]
                            : (($record[$field_key] == '*' || $record[$field_key] == '') ? $field_default : $record[$field_key]);
                    } else {
                        $record[$field_key] = $field_default;
                    }
                }

                // prepare list of zipcodes
                if (isset($record['dest_zip'])) {
                    $record['dest_zip'] = (($record['dest_zip'] == '*' || $record['dest_zip'] == '') ? '' : $record['dest_zip']);
                } else {
                    $record['dest_zip'] = '';
                }
                $record['dest_zip'] = preg_replace('/[\r\n,]+/', '{{separator}}', $record['dest_zip']);
                $record['dest_zip'] = explode('{{separator}}', $record['dest_zip']);

                // get group rules
                if (!isset($record['dest_zip_before']))
                    $record['dest_zip_before'] = '';
                else
                    $record['dest_zip_before'] = (($record['dest_zip_before'] == '*' || $record['dest_zip_before'] == '') ? '' : $record['dest_zip_before']);
                $rules = $this->getRules($record['id_carrier'], $record['id_zone'], $record['id_country'],
                    $record['id_state'], $record['dest_city'], $record['dest_zip_before'], null, null, '', $record['id_shop']);
                $group_zips = explode(',', $record['dest_zip_before']);

                // prepare delete rules with `group zips` not in `dest zips sent`
                $delete_zips = [];
                if ($action == 'edit') {
                    $delete_zips = array_diff($group_zips, $record['dest_zip']);
                    if (count($delete_zips) > 0) {
                        $group_zips = array_diff($group_zips, $delete_zips);
                    }
                }

                // insert `all rules in database` for `dest zips sent`
                if (is_array($rules) && count($rules) > 0) {
                    foreach ($rules as $rule) {
                        foreach ($record['dest_zip'] as $dest_zip) {
                            // if dest zip is in database continue
                            if (in_array($dest_zip, $group_zips)) {
                                continue;
                            }

                            // setup object
                            $carrier_table_rate = $this->carrierTableRateFactory->create();
                            foreach ($record as $record_key => $record_value) {
                                if (array_key_exists($record_key, $fields)) {
                                    if (is_int($fields[$record_key])) {
                                        $carrier_table_rate->setData($record_key, (float)$record_value);
                                    } else {
                                        $carrier_table_rate->setData($record_key, TRSHelper::addslashes($record_value));
                                    }
                                }
                            }

                            // assign rule data
                            foreach ($condition_types as $condition_type) {
                                $carrier_table_rate->setData('condition_' . $condition_type . '_from',
                                    (($rule['condition_' . $condition_type . '_from'] == '*' || $rule['condition_' . $condition_type . '_from'] == '')
                                        ? -1 : $rule['condition_' . $condition_type . '_from']));
                                $carrier_table_rate->setData('condition_' . $condition_type . '_to',
                                    (($rule['condition_' . $condition_type . '_to'] == '*' || $rule['condition_' . $condition_type . '_to'] == '')
                                        ? -1 : $rule['condition_' . $condition_type . '_to']));
                            }
                            $carrier_table_rate->setData('price', $rule['price']);
                            $carrier_table_rate->setData('comment', $rule['comment']);
                            $carrier_table_rate->setData('active', $rule['active']);

                            // assign dest zip
                            $carrier_table_rate->setData('dest_zip', TRSHelper::addslashes($dest_zip));

                            // validate object
                            $message = $this->validateFields($carrier_table_rate);

                            // if duplicate record found continue
                            if (strcasecmp($message, 'Duplicate table rate found') === 0) {
                                continue;
                            }

                            // save object of validated successfully
                            if ($message == 1) {
                                $insert_carrier_table_rate = array_intersect_key($carrier_table_rate->getData(), $fields);
                                $insert_carrier_table_rate_flat = implode(',', $insert_carrier_table_rate);
                                if (!in_array($insert_carrier_table_rate_flat, $insert_carrier_table_rates_flat)) {
                                    $insert_carrier_table_rates_flat[] = $insert_carrier_table_rate_flat;
                                    $insert_carrier_table_rates[] = $insert_carrier_table_rate;
                                    $status['success']++;
                                }
                            } else {
                                $status['error']++;
                                $status['message'][] = $message;
                            }
                        }
                    }
                    $group_zips = array_unique(array_merge(array_diff($group_zips, $record['dest_zip']), $record['dest_zip']));
                } else {
                    $group_zips = $record['dest_zip'];
                }

                // insert `rule sent` for `database dest zips`
                if (is_array($group_zips) && count($group_zips) > 0) {
                    foreach ($group_zips as $group_zip) {
                        // if group exists and sent rule is dummy continue
                        if (is_array($rules) && count($rules) > 0 && strcasecmp($record['comment'],
                                '{{dummy}}') === 0
                        ) {
                            continue;
                        }

                        // setup object
                        $carrier_table_rate = $this->carrierTableRateFactory->create();
                        foreach ($record as $record_key => $record_value) {
                            if (array_key_exists($record_key, $fields)) {
                                if (is_int($fields[$record_key])) {
                                    $carrier_table_rate->setData($record_key, (float)$record_value);
                                } else {
                                    $carrier_table_rate->setData($record_key, TRSHelper::addslashes($record_value));
                                }
                            }
                        }

                        // assign dest zip
                        $carrier_table_rate->setData('dest_zip', TRSHelper::addslashes($group_zip));

                        // validate object
                        $message = $this->validateFields($carrier_table_rate);

                        // if duplicate record found continue
                        if (strcasecmp($message, 'Duplicate table rate found') === 0) {
                            continue;
                        }

                        // save object of validated successfully
                        if ($message == 1) {
                            $insert_carrier_table_rate = array_intersect_key($carrier_table_rate->getData(), $fields);
                            $insert_carrier_table_rate_flat = implode(',', $insert_carrier_table_rate);
                            if (!in_array($insert_carrier_table_rate_flat, $insert_carrier_table_rates_flat)) {
                                $insert_carrier_table_rates_flat[] = $insert_carrier_table_rate_flat;
                                $insert_carrier_table_rates[] = $insert_carrier_table_rate;
                                $status['success']++;
                            }
                        } else {
                            $status['error']++;
                            $status['message'][] = $message;
                        }
                    }
                }

                // execute delete rules with `group zips` not in `dest zips sent`
                if ($action == 'edit' && $status['error'] <= 0 && count($delete_zips) > 0) {
                    $connection->delete(
                        $this->getMainTable(),
                        '   id_carrier = ' . (int)$record['id_carrier'] . '
                        AND id_zone = ' . (int)$record['id_zone'] . '
                        AND id_country = \'' . TRSHelper::addslashes($record['id_country'], true) . '\'
                        AND id_state = ' . (int)$record['id_state'] . '
                        AND dest_city = \'' . TRSHelper::addslashes($record['dest_city'], true) . '\'
                        AND dest_zip IN (\'' . implode("','", $delete_zips) . '\')
                        AND id_shop = ' . (int)$id_shop
                    );
                }

                // write import status
                $dir_var->writeFile('import/Kahanit/TableRateShipping/ImportStatus.json', '{"rowImported":' . ((int)$record_index + 1) . '}');
            }
        }

        if ($status['error'] > 0) {
            $status['status'] = 'danger';
            $status['message'] = implode(', ', array_unique($status['message']));

            return $status;
        } else {
            // insert records
            if (count($insert_carrier_table_rates) > 0) {
                $connection->insertMultiple($this->getMainTable(), $insert_carrier_table_rates);
                $connection->query('
                    UPDATE `' . $this->getMainTable() . '`
                    SET `order` = `id_carrier_table_rate` WHERE `order` = 0
                ');
            }

            // sort zones
            $sql = $connection->select();
            $sql->distinct();
            $sql->from(['ct' => $this->getMainTable()], ['ct.id_zone']);
            $sql->where('ct.id_carrier = ' . (int)$id_carrier . ' AND ct.id_shop = ' . (int)$id_shop);
            $sql->order(['ct.order ASC', 'ct.id_carrier_table_rate ASC']);
            $results = $connection->fetchAll($sql);
            $zones = [];
            foreach ($results as $result) {
                $zones[] = $result['id_zone'];
            }
            $this->updateOrder('zone', $id_carrier, null, $zones, $id_shop);

            // sort rule groups
            foreach ($zones as $id_zone) {
                $results = $this->getRuleGroups($id_carrier, $id_zone, null, null, '', $id_shop);
                $rulegroups = [];
                foreach ($results as $result) {
                    $rulegroups[] = $result['id_country'] .
                        '{{separator}}' . $result['id_state'] .
                        '{{separator}}' . $result['dest_city'] .
                        '{{separator}}' . $result['dest_zip'];
                }
                $this->updateOrder('rulegroup', $id_carrier, $id_zone, $rulegroups, $id_shop);
            }

            // sort rules only on edit
            if ($action == 'edit' && isset($record['id_zone']) && isset($record['id_country'])
                && isset($record['id_state']) && isset($record['dest_city']) && isset($group_zips)
            ) {
                $results = $this->getRules($id_carrier, $record['id_zone'], $record['id_country'],
                    $record['id_state'], $record['dest_city'], implode(',', $group_zips), null, null, '', $id_shop);
                $rules = [];
                foreach ($results as $result) {
                    $id_carrier_table_rate_explode = explode(',', $result['id_carrier_table_rate']);
                    $dest_zip_explode = explode(',', $result['dest_zip']);
                    $id_carrier_table_rate_explode_sorted = [];
                    foreach ($group_zips as $group_zip) {
                        $id_carrier_table_rate_explode_sorted[]
                            = $id_carrier_table_rate_explode[array_search($group_zip, $dest_zip_explode)];
                    }
                    $rules[] = implode(',', $id_carrier_table_rate_explode_sorted);
                }
                $this->updateOrder('rule', null, null, $rules, $id_shop);
            }

            // set success status
            $status['status'] = 'success';
            $status['message'] = 'Records saved successfully';

            return $status;
        }
    }

    public function import($file, $csv_separator = ',', $id_carrier, $id_shop)
    {
        $first_record = true;
        $valid_columns = [
            'zone',
            'country',
            'state',
            'city',
            'zip',
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'active',
            'comment'
        ];
        $column_numbers = [];
        $records = [];

        if (($handle = fopen($file, 'r')) !== false) {
            while (($data = fgetcsv($handle, 0, $csv_separator)) !== false) {
                if ($first_record) {
                    foreach ($data as $key => $value) {
                        if (in_array($value, $valid_columns)) {
                            $column_numbers[$value] = $key;
                        }
                    }

                    if (count($column_numbers) != count($valid_columns)) {
                        return [
                            'status' => 'danger',
                            'message' => 'CSV not properly formatted'
                        ];
                    }

                    $first_record = false;
                    continue;
                }

                $id_zone = ($data[$column_numbers['zone']] == '*' || $data[$column_numbers['zone']] == '')
                    ? 0 : $this->carrierTableRateZoneFactory->create()->getResource()->getIdByName($data[$column_numbers['zone']]);
                $id_country = ($data[$column_numbers['country']] == '*' || $data[$column_numbers['country']] == '')
                    ? '0' : $this->getCountryIdByIso($data[$column_numbers['country']]);
                $id_state = ($data[$column_numbers['state']] == '*' || $data[$column_numbers['state']] == '')
                    ? 0 : $this->getStateIdByIso($data[$column_numbers['state']], $id_country);

                $records[] = [
                    'id_zone' => $id_zone,
                    'id_country' => $id_country,
                    'id_state' => $id_state,
                    'dest_city' => ($data[$column_numbers['city']] == '*' || $data[$column_numbers['city']] == '')
                        ? '' : $data[$column_numbers['city']],
                    'dest_zip' => ($data[$column_numbers['zip']] == '*' || $data[$column_numbers['zip']] == '')
                        ? '' : preg_replace('/[\r\n,]+/', '{{separator}}', $data[$column_numbers['zip']]),
                    'condition_weight' => ($data[$column_numbers['condition_weight']] == '*' || $data[$column_numbers['condition_weight']] == '')
                        ? -1 : $data[$column_numbers['condition_weight']],
                    'condition_price' => ($data[$column_numbers['condition_price']] == '*' || $data[$column_numbers['condition_price']] == '')
                        ? -1 : $data[$column_numbers['condition_price']],
                    'condition_ptprice' => ($data[$column_numbers['condition_ptprice']] == '*' || $data[$column_numbers['condition_ptprice']] == '')
                        ? -1 : $data[$column_numbers['condition_ptprice']],
                    'condition_quantity' => ($data[$column_numbers['condition_quantity']] == '*' || $data[$column_numbers['condition_quantity']] == '')
                        ? -1 : $data[$column_numbers['condition_quantity']],
                    'condition_volume' => ($data[$column_numbers['condition_volume']] == '*' || $data[$column_numbers['condition_volume']] == '')
                        ? -1 : $data[$column_numbers['condition_volume']],
                    'price' => $data[$column_numbers['price']],
                    'active' => $data[$column_numbers['active']],
                    'comment' => $data[$column_numbers['comment']]
                ];
            }

            fclose($handle);
        }

        return $this->saveRules('add', $id_carrier, $records, $id_shop);
    }

    public function export($id_carrier, $csv_separator = ',', $id_shop)
    {
        $collection = [];
        $rule_groups = $this->getRuleGroups($id_carrier, null, null, null, '', $id_shop);
        foreach ($rule_groups as $rule_group) {
            $rules = $this->getRules($id_carrier, $rule_group['id_zone'], $rule_group['id_country'],
                $rule_group['id_state'], $rule_group['dest_city'], $rule_group['dest_zip'], null, null, '', $id_shop);

            foreach ($rules as $rule) {
                $weight = ($rule['condition_weight_from'] == $rule['condition_weight_to']) ? $rule['condition_weight_from'] : $rule['condition_weight_from'] . '-' . $rule['condition_weight_to'];
                $price = ($rule['condition_price_from'] == $rule['condition_price_to']) ? $rule['condition_price_from'] : $rule['condition_price_from'] . '-' . $rule['condition_price_to'];
                $ptprice = ($rule['condition_ptprice_from'] == $rule['condition_ptprice_to']) ? $rule['condition_ptprice_from'] : $rule['condition_ptprice_from'] . '-' . $rule['condition_ptprice_to'];
                $quantity = ($rule['condition_quantity_from'] == $rule['condition_quantity_to']) ? $rule['condition_quantity_from'] : $rule['condition_quantity_from'] . '-' . $rule['condition_quantity_to'];
                $volume = ($rule['condition_volume_from'] == $rule['condition_volume_to']) ? $rule['condition_volume_from'] : $rule['condition_volume_from'] . '-' . $rule['condition_volume_to'];

                $collection[] = (object)[
                    'zone' => ($rule_group['zone'] == '') ? '*' : $rule_group['zone'],
                    'country' => ($rule_group['c_iso_code'] == '') ? '*' : $rule_group['c_iso_code'],
                    'state' => ($rule_group['s_iso_code'] == '') ? '*' : $rule_group['s_iso_code'],
                    'city' => ($rule_group['dest_city'] == '') ? '*' : $rule_group['dest_city'],
                    'zip' => ($rule_group['dest_zip'] == '') ? '*' : implode(PHP_EOL, explode(',', $rule_group['dest_zip'])),
                    'condition_weight' => ($weight == -1) ? '*' : $weight,
                    'condition_price' => ($price == -1) ? '*' : $price,
                    'condition_ptprice' => ($ptprice == -1) ? '*' : $ptprice,
                    'condition_quantity' => ($quantity == -1) ? '*' : $quantity,
                    'condition_volume' => ($volume == -1) ? '*' : $volume,
                    'price' => $rule['price'],
                    'comment' => $rule['comment'],
                    'active' => $rule['active']
                ];
            }
        }

        if (count($collection) > 0) {
            $csv = new TRSCSV($collection, 'export', $csv_separator);
            $csv->export();
        } else {
            echo 'No records found';
        }

        die();
    }

    /* functions group four */

    public function updateCountryStateCity($update, $id_carrier, $id_zone, $id_country, $id_state, $dest_city, $dest_zip, $value, $id_shop)
    {
        // set new data variables
        if ($update == 'id_country') {
            $id_country_new = $value;
        } else {
            $id_country_new = $id_country;
        }
        if ($update == 'id_state') {
            $id_state_new = $value;
        } else {
            $id_state_new = $id_state;
        }
        if ($update == 'dest_city') {
            $dest_city_new = $value;
        } else {
            $dest_city_new = $dest_city;
        }

        // reset state of country changes
        if ($update == 'id_country') {
            $id_state_new = 0;
        }

        if ($update !== false && in_array($update, ['id_country', 'id_state', 'dest_city'])
            && $id_carrier !== 0
            && $id_zone !== -1
            && $id_country_new !== false
            && $id_state_new !== false
            && $dest_city_new !== false
        ) {
            // validate group already exists
            $rules = $this->getRules($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $dest_zip, null, null, '', $id_shop);
            foreach ($rules as $rule) {
                $id_carrier_table_rate_explode = array_map('intval', explode(',', $rule['id_carrier_table_rate']));
                foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                    $carrier_table_rate = $this->carrierTableRateFactory->create()->load($id_carrier_table_rate_explode_item);
                    $carrier_table_rate->setIdCountry(TRSHelper::addslashes((($id_country_new == '*' || $id_country_new == '') ? '0' : $id_country_new)));
                    $carrier_table_rate->setIdState((int)(($id_state_new == '*' || $id_state_new == '') ? 0 : $id_state_new));
                    $carrier_table_rate->setDestCity(TRSHelper::addslashes(($dest_city_new == '*' || $dest_city_new == '') ? '' : $dest_city_new));
                    $message = $this->validateFields($carrier_table_rate);
                    if ($message != 1) {
                        return [
                            'status' => 'danger',
                            'message' => $message
                        ];
                    }
                }
            }

            // update values
            $connection = $this->getConnection();
            $connection->update(
                $this->getMainTable(), [
                'id_country' => TRSHelper::addslashes((($id_country_new == '*' || $id_country_new == '') ? '0' : $id_country_new)),
                'id_state' => (int)(($id_state_new == '*' || $id_state_new == '') ? 0 : $id_state_new),
                'dest_city' => addslashes(($dest_city_new == '*' || $dest_city_new == '') ? '' : $dest_city_new)
            ], '    id_carrier = ' . (int)$id_carrier . '
                AND id_zone = ' . (int)$id_zone . '
                AND id_country = \'' . TRSHelper::addslashes((($id_country == '*' || $id_country == '') ? '0' : $id_country), true) . '\'
                AND id_state = ' . (int)(($id_state == '*' || $id_state == '') ? 0 : $id_state) . '
                AND dest_city = \'' . TRSHelper::addslashes((($dest_city == '*' || $dest_city == '') ? '' : $dest_city), true) . '\'
                ' . (($dest_zip == '*' || $dest_zip == '')
                    ? 'AND `dest_zip` = \'\''
                    : 'AND FIND_IN_SET(`dest_zip`, \'' . TRSHelper::addslashes($dest_zip, true) . '\')') . '
                AND id_shop = ' . (int)$id_shop
            );

            if ((int)$connection->fetchOne('SELECT ROW_COUNT()') > 0) {
                $country = $this->countryFactory->create()->load($id_country_new);
                $state = $this->regionFactory->create()->load($id_state_new);
                return [
                    'status' => 'success',
                    'id_shop' => $id_shop,
                    'id_carrier' => $id_carrier,
                    'id_zone' => $id_zone,
                    'id_country' => ($id_country_new == '0') ? '*' : $id_country_new,
                    'country' => ($country->getName() && $country->getData('iso2_code')) ? $country->getName() . ' (' . $country->getData('iso2_code') . ')' : '*',
                    'id_state' => ($id_state_new == 0) ? '*' : $id_state_new,
                    'state' => ($state->getName() && $state->getData('code')) ? $state->getName() . ' (' . $state->getData('code') . ')' : '*',
                    'dest_city' => (($dest_city_new == '') ? '*' : $dest_city_new),
                    'dest_zip' => (($dest_zip == '') ? '*' : $dest_zip),
                    'message' => 'Record updated successfully'
                ];
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateConditionPriceComment($update, $id_carrier_table_rate, $value)
    {
        $fields = [
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'comment'
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];

        if ($update !== false && in_array($update, $fields) && $id_carrier_table_rate !== false) {
            if ($update == 'price') {
                $update_values = [
                    $update => TRSHelper::addslashes($value)
                ];
            } elseif ($update == 'comment') {
                $update_values = [
                    $update => TRSHelper::addslashes($value)
                ];
            } else {
                if (is_numeric($value)) {
                    $value_from = $value_to = $value;
                } else {
                    $value_range = explode('-', $value);
                    if (isset($value_range[0]) && is_numeric($value_range[0]) && isset($value_range[1]) && is_numeric($value_range[1])) {
                        $value_from = $value_range[0];
                        $value_to = $value_range[1];
                    } else {
                        $value_from = $value_to = -1;
                    }
                }
                $update_values = [
                    $update . '_from' => (float)(($value_from == '*' || $value_from == '') ? -1 : $value_from),
                    $update . '_to' => (float)(($value_to == '*' || $value_to == '') ? -1 : $value_to)
                ];
            }

            // validate records
            $id_carrier_table_rate_explode = array_map('intval', explode(',', $id_carrier_table_rate));
            foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                $carrier_table_rate = $this->carrierTableRateFactory->create()->load($id_carrier_table_rate_explode_item);
                foreach ($update_values as $update_value_key => $update_value_value) {
                    $carrier_table_rate->setData($update_value_key, $update_value_value);
                }
                $message = $this->validateFields($carrier_table_rate);
                if ($message != 1) {
                    return [
                        'status' => 'danger',
                        'message' => $message
                    ];
                }
            }

            // update values
            $connection = $this->getConnection();
            $connection->update(
                $this->getMainTable(), $update_values,
                'id_carrier_table_rate IN (' . implode(',', array_map('intval', explode(',', $id_carrier_table_rate))) . ')'
            );

            if ((int)$connection->fetchOne('SELECT ROW_COUNT()') > 0) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                if (isset($id_carrier_table_rate[0])) {
                    $carrier_table_rate = $this->carrierTableRateFactory->create()->load((int)$id_carrier_table_rate[0]);
                    $status = [
                        'status' => 'success',
                        'price' => TRSHelper::stripslashes($carrier_table_rate->getPrice()),
                        'comment' => TRSHelper::stripslashes($carrier_table_rate->getComment()),
                        'message' => 'Record updated successfully'
                    ];
                    foreach ($condition_types as $condition_type) {
                        $status['condition_' . $condition_type . '_from'] = ($carrier_table_rate->getData('condition_' . $condition_type . '_from') == -1)
                            ? '*' : $carrier_table_rate->getData('condition_' . $condition_type . '_from');
                        $status['condition_' . $condition_type . '_to'] = ($carrier_table_rate->getData('condition_' . $condition_type . '_to') == -1)
                            ? '*' : $carrier_table_rate->getData('condition_' . $condition_type . '_to');
                    }
                    return $status;
                }
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateStatus($id_carrier_table_rate, $active)
    {
        if ($id_carrier_table_rate !== false) {
            $this->getConnection()->update(
                $this->getMainTable(), ['active' => (int)$active],
                'id_carrier_table_rate IN (' . implode(',', array_map('intval', explode(',', $id_carrier_table_rate))) . ')'
            );

            return [
                'status' => 'success',
                'message' => 'Status updated successfully',
                'id_carrier_table_rate' => $id_carrier_table_rate,
                'active' => $active
            ];
        }

        return [
            'status' => 'danger',
            'message' => 'Status update failed',
            'id_carrier_table_rate' => $id_carrier_table_rate,
            'active' => $active
        ];
    }

    public function updateOrder($entity = '', $id_carrier = null, $id_zone = null, $order = [], $id_shop)
    {
        $connection = $this->getConnection();

        if ($entity !== '' && $order !== []) {
            switch ($entity) {
                case 'zone':
                    $order_ids = [];
                    foreach ($order as $order_item) {
                        $order_ids[] = (int)preg_replace('/^zone-/', '', $order_item);
                    }
                    $sql_rank = 'SET @rank := 0';
                    $sql = 'UPDATE `' . $this->getMainTable() . '`
                        SET `order` = @rank := @rank + 1
                        WHERE `id_carrier` = ' . (int)$id_carrier . ' AND `id_shop` = ' . (int)$id_shop . '
                        ORDER BY FIELD(`id_zone`, ' . implode(',', $order_ids) . '), `order` ASC';
                    $connection->query($sql_rank);
                    $connection->query($sql);
                    break;
                case 'rulegroup':
                    if ($id_carrier !== 0 && $id_zone !== -1) {
                        $where = '`id_carrier` = ' . (int)$id_carrier . ' AND `id_zone` = ' . (int)$id_zone . ' AND `id_shop` = ' . (int)$id_shop . ' AND (';
                        foreach ($order as $order_item) {
                            $order_item_explode = explode('{{separator}}', $order_item);
                            $where .= '(`id_country` = \'' . TRSHelper::addslashes((($order_item_explode[0] == '*' || $order_item_explode[0] == '') ? '0' : $order_item_explode[0]), true) . '\'
                                AND `id_state` = ' . (int)(($order_item_explode[1] == '*' || $order_item_explode[1] == '') ? 0 : $order_item_explode[1]) . '
                                AND `dest_city` = \'' . TRSHelper::addslashes((($order_item_explode[2] == '*' || $order_item_explode[2] == '') ? '' : $order_item_explode[2]), true) . '\'
                                ' . (($order_item_explode[3] == '*' || $order_item_explode[3] == '')
                                    ? 'AND `dest_zip` = \'\''
                                    : 'AND FIND_IN_SET(`dest_zip`, \'' . TRSHelper::addslashes($order_item_explode[3], true) . '\')') . ') OR ';
                        }
                        $where = preg_replace('/ OR $/', '', $where);
                        $where .= ')';

                        // get min order rank
                        $sql = $connection->select()
                            ->from($this->getMainTable(), ['MIN(`order`) AS `minorder`'])
                            ->where($where);
                        $minorder = (int)$connection->fetchOne($sql);
                        if (!$minorder) {
                            $minorder = 1;
                        }
                        $minorder--;

                        // update order
                        foreach ($order as $order_item) {
                            $order_item_explode = explode('{{separator}}', $order_item);

                            // prepare where clause
                            $where = '`id_carrier` = ' . (int)$id_carrier . '
                                AND `id_zone` = ' . (int)$id_zone . '
                                AND `id_country` = \'' . TRSHelper::addslashes((($order_item_explode[0] == '*' || $order_item_explode[0] == '') ? '0' : $order_item_explode[0]), true) . '\'
                                AND `id_state` = ' . (int)(($order_item_explode[1] == '*' || $order_item_explode[1] == '') ? 0 : $order_item_explode[1]) . '
                                AND `dest_city` = \'' . TRSHelper::addslashes((($order_item_explode[2] == '*' || $order_item_explode[2] == '') ? '' : $order_item_explode[2]), true) . '\'
                                ' . (($order_item_explode[3] == '*' || $order_item_explode[3] == '')
                                    ? 'AND `dest_zip` = \'\''
                                    : 'AND FIND_IN_SET(`dest_zip`, \'' . TRSHelper::addslashes($order_item_explode[3], true) . '\')') . '
                                AND `id_shop` = ' . (int)$id_shop;

                            // update order query
                            $sql_rank = 'SET @rank := ' . (int)$minorder;
                            $sql = 'UPDATE `' . $this->getMainTable() . '`
                                SET `order` = @rank := @rank + 1
                                WHERE ' . $where . '
                                ORDER BY `order` ASC';
                            $connection->query($sql_rank);
                            $connection->query($sql);

                            // get no of rows updated
                            $sql = $connection->select()
                                ->from($this->getMainTable(), ['COUNT(`id_carrier_table_rate`) AS `total`'])
                                ->where($where);
                            $row_count = (int)$connection->fetchOne($sql);

                            $minorder += $row_count;
                        }
                    }
                    break;
                case 'rule':
                    $order_item_count = count(explode(',', reset($order)));
                    $order_temp = [];
                    for ($i = 0; $i < $order_item_count; $i++) {
                        foreach ($order as $order_item) {
                            $order_item_explode = array_map('intval', explode(',', $order_item));
                            $order_temp[] = $order_item_explode[$i];
                        }
                    }
                    $id_carrier_table_rate = implode(',', $order_temp);

                    // get min order rank
                    $sql = $connection->select()
                        ->from($this->getMainTable(), ['MIN(`order`) AS `minorder`'])
                        ->where('`id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')');
                    $minorder = (int)$connection->fetchOne($sql);
                    if (!$minorder) {
                        $minorder = 1;
                    }
                    $minorder--;

                    // update order
                    $sql_rank = 'SET @rank := ' . (int)$minorder;
                    $sql = 'UPDATE `' . $this->getMainTable() . '`
                        SET `order` = @rank := @rank + 1
                        WHERE `id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')
                        ORDER BY FIELD(`id_carrier_table_rate`, ' . $id_carrier_table_rate . '), `order` ASC';
                    $connection->query($sql_rank);
                    $connection->query($sql);
                    break;
            }
        }

        return [
            'status' => 'success',
            'message' => 'Order saved successfully'
        ];
    }

    /* functions group five */

    public function deleteRules($entity, $type, $id_carrier, $data, $id_zone, $id_carrier_table_rate, $id_shop)
    {
        $connection = $this->getConnection();

        if ($entity == 'rule') {
            if ($id_carrier_table_rate !== false && ($type == 'single' || $type == 'selected')) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                $id_carrier_table_rate = implode(',', $id_carrier_table_rate);
                $connection->delete($this->getMainTable(),
                    'id_carrier_table_rate IN (' . $id_carrier_table_rate . ')');
            }
        } elseif ($entity == 'rulegroup') {
            if ($data !== false && ($type == 'single' || $type == 'selected')) {
                foreach ($data as $line) {
                    $connection->delete($this->getMainTable(),
                        '   id_carrier = ' . (int)$id_carrier . '
                        AND id_zone = ' . (int)$line['id_zone'] . '
                        AND id_country = \'' . TRSHelper::addslashes((($line['id_country'] == '*' || $line['id_country'] == '') ? '0' : $line['id_country']), true) . '\'
                        AND id_state = ' . (int)(($line['id_state'] == '*' || $line['id_state'] == '') ? 0 : $line['id_state']) . '
                        AND dest_city = \'' . TRSHelper::addslashes((($line['dest_city'] == '*' || $line['dest_city'] == '') ? '' : $line['dest_city']), true) . '\'
                        ' . (($line['dest_zip'] == '*' || $line['dest_zip'] == '')
                            ? 'AND `dest_zip` = \'\''
                            : 'AND FIND_IN_SET(`dest_zip`, \'' . TRSHelper::addslashes($line['dest_zip'], true) . '\')') . '
                        AND id_shop = ' . (int)$id_shop);
                }
            } elseif ($id_zone !== -1 && $type == 'all') {
                $connection->delete($this->getMainTable(),
                    'id_zone = ' . (int)$id_zone . ' AND id_shop = ' . (int)$id_shop);
            }
        }

        return [
            'status' => 'success',
            'message' => 'Table rates deleted successfully'
        ];
    }

    public function deleteRulesByCarrier($id_carrier)
    {
        $this->getConnection()->delete($this->getMainTable(), 'id_carrier = ' . (int)$id_carrier);
    }

    public function deleteRulesByZone($id_zone)
    {
        $this->getConnection()->delete($this->getMainTable(), 'id_zone = ' . (int)$id_zone);
    }

    /* functions group six */

    public function getZonesOrder($id_carrier, $id_shop)
    {
        $connection = $this->getConnection();
        $order = [];

        if ($id_carrier !== 0) {
            $sql = $connection->select();
            $sql->distinct();
            $sql->from(['ct' => $this->getMainTable()], ['ct.id_zone']);
            $sql->where('ct.id_carrier = ' . (int)$id_carrier . ' AND ct.id_shop = ' . (int)$id_shop);
            $sql->order(['ct.order ASC', 'ct.id_carrier_table_rate ASC']);
            $results = $connection->fetchAll($sql);
            foreach ($results as $result) {
                $order[] = $result['id_zone'];
            }
        }

        return $order;
    }

    public function getCountries($search, $id_shop)
    {
        $collection = $this->countryFactory->create()->getCollection();

        $results = [];
        foreach ($collection->getItems() as $key => $value) {
            $name = $value->getName();
            if ($search == '' || stristr($name, $search) !== false) {
                $results[] = [
                    'id' => $value->getId(),
                    'text' => $name
                ];
            }
        }

        return $results;
    }

    public function getStates($search, $id_country, $id_shop)
    {
        $collection = $this->regionFactory->create()->getCollection()
            ->addFieldToFilter('country_id', ['=' => $id_country])
            ->addFieldToFilter('name', ['like' => '%' . TRSHelper::addslashes($search, true) . '%']);

        $results = [];
        foreach ($collection->getItems() as $key => $value) {
            $results[] = [
                'id' => $value->getId(),
                'text' => $value->getName()
            ];
        }

        return $results;
    }

    public function getCountryIdByIso($iso)
    {
        $collection = $this->countryFactory->create()->getCollection()
            ->addFieldToFilter('iso2_code', ['=', $iso]);
        $country = $collection->getFirstItem();

        return $country->getId();
    }

    public function getStateIdByIso($iso, $id_country)
    {
        $collection = $this->regionFactory->create()->getCollection()
            ->addFieldToFilter('code', ['=', $iso])
            ->addFieldToFilter('country_id', ['=', $id_country]);
        $state = $collection->getFirstItem();

        return $state->getId();
    }

    /* functions group six */

    public function getShippingPrice($id_carrier, $id_country, $id_state, $dest_city, $dest_zip, $vars, $id_shop)
    {
        if (!class_exists('Symfony\\Component\\ExpressionLanguage\\ExpressionLanguage')) {
            return false;
        }

        $dest_zip = str_replace('-', '', str_replace(' ', '', $dest_zip));

        $connection = $this->getConnection();
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['ct.*', 'SUBSTRING_INDEX(ct.dest_zip, "-", 1) AS dest_zip_from', 'SUBSTRING_INDEX(ct.dest_zip, "-", -1) AS dest_zip_to']);
        $sql->where('
                (ct.id_shop     = ' . (int)$id_shop . ' OR ct.id_shop = ' . Store::DEFAULT_STORE_ID . ')
            AND ct.id_carrier  = ' . (int)$id_carrier . '
            AND (ct.id_country = \'' . TRSHelper::addslashes($id_country, true) . '\' OR ct.id_country = \'0\')
            AND (ct.id_state   = ' . (int)$id_state . ' OR ct.id_state   = 0)
            AND (ct.dest_city  = \'' . TRSHelper::addslashes($dest_city, true) . '\' OR ct.dest_city  = "")
            AND (' . (float)$vars['tw'] . '   BETWEEN ct.condition_weight_from   AND ct.condition_weight_to   OR (ct.condition_weight_from   < 0 AND ct.condition_weight_to   < 0))
            AND (' . (float)$vars['tp'] . '   BETWEEN ct.condition_price_from    AND ct.condition_price_to    OR (ct.condition_price_from    < 0 AND ct.condition_price_to    < 0))
            AND (' . (float)$vars['tptp'] . ' BETWEEN ct.condition_ptprice_from  AND ct.condition_ptprice_to  OR (ct.condition_ptprice_from  < 0 AND ct.condition_ptprice_to  < 0))
            AND (' . (float)$vars['tq'] . '   BETWEEN ct.condition_quantity_from AND ct.condition_quantity_to OR (ct.condition_quantity_from < 0 AND ct.condition_quantity_to < 0))
            AND (' . (float)$vars['tv'] . '   BETWEEN ct.condition_volume_from   AND ct.condition_volume_to   OR (ct.condition_volume_from   < 0 AND ct.condition_volume_to   < 0))
            AND ct.active = 1
        ');
        $sql->having('
            (
                dest_zip_from = ""
                AND dest_zip_to = ""
            )
            OR (
                dest_zip_from = dest_zip_to
                AND \'' . TRSHelper::addslashes($dest_zip, true) . '\' LIKE REPLACE(REPLACE(REPLACE(dest_zip_to, "*", "%"), " ", ""), "-", "")
            )
            OR (
                dest_zip_from = dest_zip_to
                AND \'' . TRSHelper::addslashes($dest_zip, true) . '\' LIKE REPLACE(REPLACE(REPLACE(dest_zip_from, "*", "%"), " ", ""), "-", "")
            )
            OR (
                dest_zip_from != ""
                AND dest_zip_to != ""
                AND REPLACE(REPLACE(dest_zip_from, " ", ""), "-", "") <= \'' . TRSHelper::addslashes($dest_zip, true) . '\'
                AND REPLACE(REPLACE(dest_zip_to, " ", ""), "-", "") >= \'' . TRSHelper::addslashes($dest_zip, true) . '\'
            )
        ');
        $sql->order(['FIELD(`id_shop`, ' . $id_shop . ', ' . Store::DEFAULT_STORE_ID . ')', 'ct.order ASC', 'ct.id_carrier_table_rate ASC']);
        $results = $connection->fetchAll($sql);

        $language = new ExpressionLanguage();
        $language->registerProvider(new TRSExpressionFunctionProvider());
        $shipping_price = false;

        foreach ($results as $result) {
            $vars_temp = $vars;
            $vars_temp = array_merge($vars_temp, [
                'ctwf' => $result['condition_weight_from'],
                'ctwt' => $result['condition_weight_to'],
                'ctwi' => $vars_temp['tw'] - $result['condition_weight_from'],
                'ctpf' => $result['condition_price_from'],
                'ctpt' => $result['condition_price_to'],
                'ctpi' => $vars_temp['tp'] - $result['condition_price_from'],
                'ctptpf' => $result['condition_ptprice_from'],
                'ctptpt' => $result['condition_ptprice_to'],
                'ctptpi' => $vars_temp['tptp'] - $result['condition_ptprice_from'],
                'ctqf' => $result['condition_quantity_from'],
                'ctqt' => $result['condition_quantity_to'],
                'ctqi' => $vars_temp['tq'] - $result['condition_quantity_from'],
                'ctvf' => $result['condition_volume_from'],
                'ctvt' => $result['condition_volume_to'],
                'ctvi' => $vars_temp['tv'] - $result['condition_volume_from']
            ]);

            $result['price'] = str_replace('$', '', $result['price']); // for bc with old formula variables with $

            try {
                $shipping_price = $language->evaluate($result['price'], $vars_temp);
            } catch (\Exception $e) {
                $shipping_price = false;
            }

            if ($shipping_price < 0) {
                continue;
            } else {
                break;
            }
        }

        return $shipping_price;
    }
}
